@@ -206,24 +206,133 @@ static void text_output_footer(void)
     puts("");
 }
 
+static char *xml_escaped(const char *const s_) {
+    static char buf[MAXPATHLEN + 32U];
+    const unsigned char *s = (const unsigned char *) s_;
+    char *bufpnt = buf;
+    size_t left = sizeof buf - (size_t) 1U;
+    
+    while (left > (size_t) 0U && *s != 0U) {
+        if (ISCTRLCODE(*s)) {
+            if (left <= (size_t) 0U) {
+                *bufpnt = 0;
+                return buf;
+            }
+            *bufpnt++ = '?';
+            left--;
+            goto next;
+        }
+        switch (*s) {
+        case '<':
+            if (left < sizeof "&lt;" - (size_t) 1U) {
+                *bufpnt = 0;
+                return buf;
+            }
+            *bufpnt++ = '&'; *bufpnt++ = 'l'; *bufpnt++ = 't'; *bufpnt++ = ';';
+            left -= sizeof "&lt;" - (size_t) 1U;
+            break;
+        case '>':
+            if (left < sizeof "&gt;" - (size_t) 1U) {
+                *bufpnt = 0;
+                return buf;
+            }
+            *bufpnt++ = '&'; *bufpnt++ = 'g'; *bufpnt++ = 't'; *bufpnt++ = ';';
+            left -= sizeof "&gt;" - (size_t) 1U;
+            break;
+        case '&':
+            if (left < sizeof "&amp;" - (size_t) 1U) {
+                *bufpnt = 0;
+                return buf;
+            }
+            *bufpnt++ = '&'; *bufpnt++ = 'a'; *bufpnt++ = 'm'; *bufpnt++ = 'p';
+            *bufpnt++ = ';';
+            left -= sizeof "&amp;" - (size_t) 1U;
+            break;
+        default:
+            *bufpnt++ = (char) *s;
+            left--;
+        }
+        next:
+        s++;
+    }
+    *bufpnt = 0;
+    
+    return buf;    
+}
+
 /* HTML output */
 
+static char *html_escaped(const char *const s_) {
+    return xml_escaped(s_);
+}
+
 static void html_output_header(void)
 {
     if (html_cgi != 0) {
         puts("Content-Type: text/html\n");
     }
     if (html_raw == 0) {
-        puts("<!DOCTYPE html PUBLIC \"-//W3C/DTD XHTML 1.0 Transitional//EN\" "
-             "\"DTD/xhtml1-transitional.dtd\">\n"
-             "<html>\n"
-             "<title>Pure-FTPd server status</title>\n"
-             "<body bgcolor=\"#ffffff\" text=\"#000000\">\n");
-        puts("<div align=\"center\">\n");
+        puts("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"\n"
+             "          \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n"
+             "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">\n"
+             "<head>\n"             
+             " <meta http-equiv=\"Content-Type\"\n" 
+             "       content=\"text/html; charset=ISO-8859-15\" />\n"
+             " <title>Pure-FTPd server status</title>\n"
+             " <style type=\"text/css\">\n"
+             "html {\n"
+             " background-color: #369;\n"
+             "}\n"
+             "body {\n"
+             " background-color: #fff;\n"
+             " color: #000;\n"
+             " margin: 12px;\n"
+             " padding: 8px;\n"
+             " border: 2px solid #000;\n"
+             " font-family: \"Trebuchet MS\",Verdana,Geneva,Arial,Helvetica,sans-serif;\n"
+             " font-size: 0.8em;\n"
+             "}\n"
+             "h1 {\n"
+             " text-align: center;\n"
+             " border-bottom: 1px solid #666;\n"
+             " margin: 0.5em 1em;\n"
+             "}\n"
+             "#ftp-status {\n"
+             " text-align: center;\n"
+             "}\n"
+             "table {\n"
+             " margin: 0 auto;\n"
+             "}\n"
+             "thead th {\n"
+             " background-color: #369;\n"
+             " color: #fff;\n"
+             "}\n"
+             "th,td {\n"
+             " padding: 0.1em 0.5em;\n"
+             "}\n"
+             "tr:hover {\n"
+             " background-color: #def;\n"
+             "}\n"
+             " </style>\n"
+             "</head>\n"
+             "<body>\n"
+             "<h1>Pure-FTPd server status</h1>");
     }
-    puts("<table width=\"100%\" cellspacing=\"4\" border=\"2\" cellpadding=\"4\">");
-    puts("<th>PID</th><th>Account</th><th>Time</th><th>State</th>"
-         "<th>File name</th><th>Remote host</th><th>Kbytes</th><th>Local host</th>");
+    puts("<div id=\"ftp-status\">\n"
+         " <table summary=\"Pure-FTPd server status\">\n"
+         "  <thead>\n"
+         "   <tr>\n"
+         "    <th scope=\"col\">PID</th>\n"
+         "    <th scope=\"col\">Account</th>\n"
+         "    <th scope=\"col\">Time</th>\n"
+         "    <th scope=\"col\">State</th>\n"
+         "    <th scope=\"col\" abbr=\"File\">File name</th>\n"
+         "    <th scope=\"col\" abbr=\"Peer\">Remote host</th>\n"
+         "    <th scope=\"col\" abbr=\"Kb\">Kbytes</th>\n"
+         "    <th scope=\"col\" abbr=\"Local\">Local host</th>\n"
+         "   </tr>\n"
+         "  </thead>\n"
+         "  <tbody>");
 }
 
 static void html_output_line(const pid_t pid, const char * const account,
@@ -238,18 +347,16 @@ static void html_output_line(const pid_t pid, const char * const account,
                              const off_t total_size,
                              const off_t current_size)
 {    
-    puts("<tr valign=\"middle\">");
-    printf("<td>%lu</td>"
-           "<td>%s</td>"
-           "<td>%02lu:%02lu</td>"
-           "<td>%s</td>"
-           "<td>%s</td>"
-           "<td>%s</td>\n",
-           (unsigned long) pid, 
-           *account == 0 ? "&nbsp;" : account,
-           (since / 60UL) / 60UL, (since / 60UL) % 60UL, state, 
-           *filename == 0 ? "&nbsp;" : filename, 
-           hbuf);
+    puts("   <tr>");
+    printf("    <th scope=\"row\">%lu</th>\n", (unsigned long) pid);
+    printf("    <td>%s</td>\n",
+           *account == 0 ? "&nbsp;" : html_escaped(account));           
+    printf("    <td>%02lu:%02lu</td>\n",
+           (since / 60UL) / 60UL, (since / 60UL) % 60UL);
+    printf("    <td>%s</td>\n", html_escaped(state));
+    printf("    <td>%s</td>\n",
+           *filename == 0 ? "&nbsp;" : html_escaped(filename));
+    printf("    <td>%s</td>\n", html_escaped(hbuf));
     if (current_size > (off_t) 0) {
         unsigned long bandwidth;
         
@@ -268,59 +375,39 @@ static void html_output_line(const pid_t pid, const char * const account,
             if (pcti > 100) {
                 pcti = 100;           /* should never happen */
             }
-            printf("<td>%llu/%llu (%d%% - %lu KB/s)</td>\n",
+            printf("    <td>%llu/%llu (%d%% - %lu KB/s)</td>\n",
                    (unsigned long long) (current_size / 1024),
                    (unsigned long long) (total_size / 1024), 
                    pcti, bandwidth);
         } else {
-            printf("<td>%llu (%lu KB/s)</td>\n",
+            printf("    <td>%llu (%lu KB/s)</td>\n",
                    (unsigned long long) (current_size / 1024),
                    bandwidth);
         }
     } else {
-        puts("<td>&nbsp;</td>");
+        puts("    <td>&nbsp;</td>");
     }
-    printf("<td>%s:%s</td>\n", local_hbuf, local_port);
-    puts("</tr>");
+    printf("    <td>%s:", html_escaped(local_hbuf));
+    printf("%s</td>\n", html_escaped(local_port));
+    puts("   </tr>");
 }
 
 static void html_output_footer(void)
 {
-    puts("</table>");
+    puts("  </tbody>\n"
+         " </table>\n"
+         "</div>");
     if (html_raw == 0) {
-        puts("\n</div>\n");
-        puts("\n</body>\n</html>");
+        puts("</body>\n"
+             "</html>");
     }
 }
 
 /* XML output */
 
-static const char *xml_escaped(const char * const s_)
-{
-    register const unsigned char *s = (const unsigned char *) s_;        
-    static char buf[1024];
-    register char *bufpnt = buf;    
-    
-    if (strlen(s_) >= sizeof buf) {
-        return "*";
-    }
-    while (*s != 0U) {
-        if (*s < 32U || *s > 126U || *s == '"') {
-            *bufpnt = '?';
-        } else {
-            *bufpnt = (char) *s;
-        }
-        bufpnt++;
-        s++;
-    }
-    *bufpnt = 0;
-    
-    return buf;
-}
-
 static void xml_output_header(void)
 {
-    puts("<?xml version=\"1.0\" ?>\n\n"
+    puts("<?xml version=\"1.0\" encoding=\"ISO-8859-15\"?>\n"
          "<status>");
 }
 
@@ -380,34 +467,130 @@ static void xml_output_footer(void)
     puts("</status>");
 }
 
+/* Apple / GNUStep plist output */
+
+static void plist_output_header(void)
+{
+    puts("<?xml version=\"1.0\" encoding=\"ISO-8859-15\"?>");
+    puts("<!DOCTYPE plist PUBLIC \"-//Apple Computer//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">");
+    puts("<plist version=\"1.0\">\n<dict>");
+    puts("\t<key>user-info</key>\n\t<array>");
+}
+
+static void plist_output_line(const pid_t pid, const char * const account,
+                              const unsigned long since,
+                              const unsigned long xfer_since,
+                              const char * const state,
+                              const char * const filename,
+                              const char * const hbuf,
+                              const char * const local_hbuf,
+                              const char * const local_port,
+                              const off_t restartat,                              
+                              const off_t total_size,
+                              const off_t current_size)
+{
+    puts("\t\t<dict>");
+    puts("\t\t\t<key>pid</key>");
+    printf("\t\t\t<string>%lu</string>\n", (unsigned long) pid);
+    puts("\t\t\t<key>account</key>");
+    printf("\t\t\t<string>%s</string>\n", xml_escaped(account));
+    puts("\t\t\t<key>time</key>");
+    printf("\t\t\t<string>%lu</string>\n", since);
+    puts("\t\t\t<key>state</key>");
+    printf("\t\t\t<string>%s</string>\n", state);
+    puts("\t\t\t<key>file</key>");
+    printf("\t\t\t<string>%s</string>\n", xml_escaped(filename));
+    puts("\t\t\t<key>host</key>");
+    printf("\t\t\t<string>%s</string>\n", xml_escaped(hbuf));
+    puts("\t\t\t<key>localhost</key>");
+    printf("\t\t\t<string>%s</string>\n", xml_escaped(local_hbuf));
+    puts("\t\t\t<key>localport</key>");
+    printf("\t\t\t<string>%s</string>\n", xml_escaped(local_port));
+    
+    if (current_size > (off_t) 0) {
+        unsigned long bandwidth;
+        long double pct;
+        int pcti;        
+        
+        if (xfer_since > 0UL && current_size > restartat) {
+            bandwidth = (unsigned long) ((current_size - restartat) / xfer_since);
+        } else {
+            bandwidth = 0UL;
+        }        
+        if ((long double) total_size > 0.0L) {
+            pct = ((long double) current_size * 100.0L) / (long double) total_size;
+            pcti = (int) (pct + 0.5L);
+            if (pcti > 100) {
+                pcti = 100;           /* should never happen */
+            }
+            puts("\t\t\t<key>resume</key>");
+            printf("\t\t\t<string>%llu</string>\n", (unsigned long long) restartat);
+            puts("\t\t\t<key>current_size</key>");
+            printf("\t\t\t<string>%llu</string>\n", (unsigned long long) current_size);
+            puts("\t\t\t<key>total_size</key>");
+            printf("\t\t\t<string>%llu</string>\n", (unsigned long long) total_size);
+            puts("\t\t\t<key>percentage</key>");
+            printf("\t\t\t<string>%d</string>\n", pcti);
+            puts("\t\t\t<key>bandwidth</key>");
+            printf("\t\t\t<string>%lu</string>\n", bandwidth);            
+        } else {
+            puts("\t\t\t<key>resume</key>");
+            printf("\t\t\t<string>%llu</string>\n", (unsigned long long) restartat);
+            puts("\t\t\t<key>current_size</key>");
+            printf("\t\t\t<string>%llu</string>\n", (unsigned long long) current_size);
+            puts("\t\t\t<key>bandwidth</key>");
+            printf("\t\t\t<string>%lu</string>\n", bandwidth);
+        }
+    } else {       
+        puts("\t\t\t<key>resume</key>");
+        puts("\t\t\t<string></string>");
+        puts("\t\t\t<key>current_size</key>");
+        puts("\t\t\t<string></string>");
+        puts("\t\t\t<key>total_size</key>");
+        puts("\t\t\t<string></string>");
+        puts("\t\t\t<key>percentage</key>");
+        puts("\t\t\t<string></string>");
+        puts("\t\t\t<key>bandwidth</key>");
+        puts("\t\t\t<string></string>");
+    }
+    puts("\t\t</dict>");
+}
+
+static void plist_output_footer(void)
+{
+    puts("\t</array>");
+    puts("</dict>\n</plist>");
+}
+
 /* Shell output */
 
 static const char *shell_escaped(const char * const s_)
 {
     register const unsigned char *s = (const unsigned char *) s_;        
-    static char buf[2048];
-    const char * const bufend = &buf[sizeof buf];
+    static char buf[MAXPATHLEN + 32U];
+    const char * const bufend = &buf[sizeof buf - (size_t) 1U];
     register char *bufpnt = buf;    
     
-    if (strlen(s_) >= sizeof buf) {
-        return "-";
-    }
     while (*s != 0U) {
         if (ISCTRLCODE(*s)) {
             *bufpnt = '_';
-        } else if (*s == '|') {
+        } else if (*s == '|' || *s == '\\') {
+            if (bufpnt == bufend) {
+                break;
+            }
             *bufpnt++ = '\\';
             if (bufpnt == bufend) {
-                return "-";
+                bufpnt--;
+                break;
             }
-            *bufpnt = '|';
+            *bufpnt = (char) *s;
         } else {
             *bufpnt = (char) *s;
         }
-        bufpnt++;
         if (bufpnt == bufend) {
-            return "-";
+            break;
         }
+        bufpnt++;
         s++;
     }
     *bufpnt = 0;
@@ -427,7 +610,7 @@ static void shell_output_line(const pid_t pid, const char * const account,
                               const char * const hbuf,
                               const char * const local_hbuf,
                               const char * const local_port,
-                              const off_t restartat,                              
+                              const off_t restartat,
                               const off_t total_size,
                               const off_t current_size)
 {
@@ -491,6 +674,7 @@ static void help(void)
          "-h : help\n"
          "-H : don't resolve host names, only show IP addresses\n"
          "-n : synonym for -H\n"
+         "-p : output Apple / GNUStep plist data\n"
          "-s : easily parsable output for shell scripts\n"
          "     format is :\n"
          "     pid|acct|time|state|file|peer|local|port|current|total|%|bandwidth\n"
@@ -573,12 +757,12 @@ int main(int argc, char *argv[])
 # endif
     
     if (getenv("GATEWAY_INTERFACE") != NULL) {
-        html_cgi = html_raw = 1;
+        html_cgi = 1;
         output_header = html_output_header;
         output_line = html_output_line;            
         output_footer = html_output_footer;        
     }
-    while ((fodder = getopt(argc, argv, "CchHnsvwWx")) != -1) {
+    while ((fodder = getopt(argc, argv, "CchHnpsvwWx")) != -1) {
         switch(fodder) {
         case 'h' :
             help();
@@ -597,6 +781,11 @@ int main(int argc, char *argv[])
             output_line = html_output_line;            
             output_footer = html_output_footer;
             break;
+        case 'p' :
+            output_header = plist_output_header;
+            output_line = plist_output_line;
+            output_footer = plist_output_footer;
+            break;
         case 's' :
             output_header = shell_output_header;
             output_line = shell_output_line;            
@@ -717,6 +906,11 @@ int main(int argc, char *argv[])
                     break;
                 }                
 #endif
+                goto nextone;
+            }
+            for (;;) {
+                int eai;
+                
                 if ((eai = getnameinfo
                      ((struct sockaddr *) &scanned_entry->local_addr,
                       STORAGE_LEN(scanned_entry->addr),
@@ -764,8 +958,7 @@ int main(int argc, char *argv[])
         if (delete_file != 0) {
             unlink(entry->d_name);
         }
-    }
-    
+    }    
     output_footer();
     
     return 0;
